/*
 * NS-Link support daemon
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <poll.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <syslog.h>
#include <grp.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netpacket/packet.h>
#include <net/if.h>

#include "nslink.h"
#include "nslinkd.h"
#include "version.h"

char *dfl_boot_file4_8 = "/etc/rpshsi.bin";
char *dfl_boot_file2 = "/etc/rpshsi2p.bin";
const char *not_loaded = "RPSH-Si driver not loaded?";
const char *pid_file = "/var/run/nslinkd.pid";

static pid_t child_pids[MAX_NRP_SIS];
static pid_t parent_pid;
static int child_socket;

static int get_major_number(const char *name)
{
  FILE *f;
  char buf[80], devname[80];
  int n, major = -1;

  f = fopen("/proc/devices", "r");
  if (!f)
    return -1;
  while (!feof(f))
    {
      if (!fgets(buf, sizeof(buf), f))
	break;
      if (sscanf(buf, "%d %s", &n, devname) != 2)
	continue;
      if (!strcmp(devname, name))
	{
	  major = n;
	  break;
	}
    }
  fclose(f);
  return major;
}

// using tempnam generate linker warnings under gcc 2.96, so we'll write our own :(
#if !defined(P_tmpdir)
# define P_tmpdir "/tmp"
#endif  

static char *myTempName(char *prefix)
{
  int tries;
  char *tmpdir;
  static char path[256];
  struct stat statbuf;
  
  tmpdir = getenv("TMPDIR");
  if (!tmpdir)
    tmpdir = P_tmpdir;
  if (strlen(tmpdir)+strlen(prefix)+22 >= sizeof(path))
    {
      syslog(LOG_ERR,"Temporary directory path/prefix too long: %s/%s\n",tmpdir,prefix);
      return NULL;
    }
  srandom(time(NULL));
  for (tries=0; tries<10; ++tries)
    {
      int s;
      sprintf(path,"%s/%s.%08lx.%08lx",tmpdir,prefix,(long)time(NULL),(long)random());
      s = stat(path,&statbuf);
      if (s)
        return path;
    }
  return NULL;
}

static int open_dev(dev_t dev, int mode)
{
  char *fn;
  int fd;

  if ((fn = myTempName("nslink")) == NULL)
    return -1;
  if (mknod(fn, S_IFCHR | 0600, dev) != 0)
    return -1;
  fd = open(fn, mode);
  unlink(fn);
  return fd;
}

int config_rpshsi(int fd,   struct si_box_config *config, int boxid)
{
  struct rpshsi_box_config cfg;
  int ret, i;

  //syslog(LOG_INFO, "config_rpshsi(fd=%d, boxid=%d, ...)\n",fd,boxid);  /* debug */

  memset(&cfg, 0, sizeof cfg);
  cfg.boxid = boxid;
  cfg.base_port = config->base_port;
  cfg.num_ports = config->num_lines;
  memcpy(cfg.ether_addr, config->ether_addr, 6);
  strcpy(cfg.interface, config->interface);
  cfg.ip_addr = config->ip_addr;
  cfg.link_timeout = config->link_timeout;
  cfg.scan_period = config->scan_period;
  cfg.arch_id = config->arch_id;
  for (i=0; i<MAX_NRP_PORTS; ++i)
    {
      cfg.rsmode[i] = config->rsmode[i];
      cfg.low_latency[i] = config->low_latency[i];
      cfg.inter_char_timeout[i] = config->inter_char_timeout[i];
    }
  //syslog(LOG_INFO,"boxid %d link timeout = %d seconds\n",boxid,config->link_timeout);  /* debug */
  ret = ioctl(fd, RPSHSI_SET_BOX_CONFIG, &cfg);
  if (ret)
    {
      syslog(LOG_ERR, "Couldn't config NS-Link box #%d (errno %d)",boxid,errno);
      return -1;
    }
  //syslog(LOG_INFO,"config_rpshsi return\n");  /* debug */
  return 0;
}


int become_daemon()
{
  int child_pid, fd;

  child_pid = fork();
  if (child_pid == -1)
    return -1;
  if (child_pid)
    exit(0);
  (void) chdir("/");
  fd = open("/dev/null", O_RDWR, 0);
  if (fd >= 0)
    {
      (void) dup2(fd, 0);
      (void) dup2(fd, 1);
      (void) dup2(fd, 2);
      if (fd > 2)
	(void) close(fd);
    }
  return (0);
}

void check_pid_file()
{
  FILE *f;
  int pid;

  if (access(pid_file, R_OK) == 0)
    {
      f = fopen(pid_file, "r");
      if (f)
	{
	  pid = 0;
	  fscanf(f, "%d", &pid);
	  if (pid && !kill(pid, 0))
	    {
	      fprintf(stderr, "NS-Link daemon already running!\n");
	      syslog(LOG_ERR, "NS-Link daemon already running");
	      exit(1);
	    }
	  fclose(f);
	}
    }
}

void put_pid_file()
{
  FILE *f;

  f = fopen(pid_file, "w");
  if (!f)
    {
      syslog(LOG_WARNING, "Warning: couldn't write pid file %s", pid_file);
      return;
    }
  parent_pid = getpid();
  fprintf(f, "%d\n", parent_pid);
  fclose(f);
}

/*
 * Create a device if it doesn't exist and have the correct device
 * number or major/minor number.
 */
#define MKDEV(a,b) ((int)((((a) & 0xff) << 8) | ((b) & 0xff)))


void create_device(const char *prefix, int major, int num, gid_t def_group)
{
  char name[80];
  struct stat st;
  dev_t device;
  mode_t mode = 0620;  // default file protection
  uid_t  owner = 0;    // default owner
  gid_t  group = def_group;

  sprintf(name, "%s%d", prefix, num);
  // syslog(LOG_INFO, "Device created %s\n", name);
  device = MKDEV(major, num);
  if (lstat(name, &st) == 0)
    {
      if ((st.st_rdev == device) && S_ISCHR(st.st_mode))
	return;
#if 0
      syslog(LOG_INFO, "%s has dev %d, mode %o wanted %d",
	     name, st.st_rdev, st.st_mode, device);
#endif
      // device exists, but doesn't have right major/minor numbers,
      // so remember owner/group/mode so we can create new device 
      // with same properties as old one except w/ right major/minor.
      
      mode = st.st_mode & 0777;
      owner = st.st_uid;
      group = st.st_gid;
      
      (void) unlink(name);
    }
  
  umask(0);
  
  if (mknod(name, S_IFCHR | mode, device) < 0)
    {
      perror(name);
      syslog(LOG_ERR, "Couldn't create device %s: %m", name);
    }
  if (chown(name, owner, group) < 0)
    {
      perror(name);
      syslog(LOG_ERR, "Couldn't chmod device %s: %m", name);
    }
}

void unlink_device(const char *prefix, int num)
{
  char name[80];
  sprintf(name,"%s%d",prefix,num);
  unlink(name);
}


/*
 * This function creates all of the devices
 */
void create_devices()
{
  int i, api, num_ports;
  struct group *grp;
  gid_t tty_group = 0;
  int major, callout_major;
  unsigned char active_port[(MAX_NRP_SIS * MAX_NRP_PORTS)];

  syslog(LOG_INFO, "Creating devices\n");
  major = get_major_number("ttySI");
  callout_major = get_major_number("cuSI");

  if (major == 0 && callout_major == 0)
    return;

  num_ports = 0;
  api = 0;
  for (i = 0; i < num_box_config; i++)
    {
      num_ports += box_config[i].num_lines;
      memcpy(&active_port[api], box_config[i].rsmode,
	     box_config[i].num_lines);
      api += box_config[i].num_lines;
    }

  syslog(LOG_INFO, "NumPorts: %d\n", num_ports);
  grp = getgrnam("tty");
  if (grp)
    tty_group = grp->gr_gid;

  for (i=0; i<num_ports; ++i)
    {
      if (major)
        create_device("/dev/ttySI", major, i, tty_group);
      if (callout_major)
        create_device("/dev/cuSI", callout_major, i, tty_group);
    }
}

void sigterm_handler()
{
  int i, rc;

  if (parent_pid == getpid())
    {
      for (i = 0; i < num_box_config; i++)
        if (child_pids[i])
          kill(child_pids[i], SIGTERM);
      syslog(LOG_NOTICE, "Exiting");
      (void) unlink(pid_file);
    }
  else
    {
      rc = shutdown(child_socket, 2);
      rc = close(child_socket);
    }
  exit(0);
}

void sighup_handler()
{
  signal(SIGHUP, sighup_handler);
  return;
}

int socketReadTimeout(int socket, void *buf, int count, unsigned usTimeout)
{
  fd_set r,e;
  struct timeval tv;
  int s;
  
  FD_ZERO(&r);
  FD_ZERO(&e);
  FD_SET(socket,&r);
  FD_SET(socket,&e);
  tv.tv_sec = usTimeout/1000000;
  tv.tv_usec = usTimeout % 1000000;
  
  s = select(socket+1,&r,NULL,&e,&tv);
  
  if (s==0)
    return 0;
  
  if (!FD_ISSET(socket,&r))
    return 0;
  
  return read(socket,buf,count);
}


int packetReadTimeout(int socket, unsigned char *macAddr, unsigned char index, void *b, int count,  unsigned usTimeout)
{
  unsigned char *buf = b;
  fd_set r,e;
  struct timeval tv;
  int s; 
  int n;
  
  
  FD_ZERO(&r);
  FD_ZERO(&e);
  FD_SET(socket,&r);
  FD_SET(socket,&e);
  tv.tv_sec = usTimeout/1000000;
  tv.tv_usec = usTimeout % 1000000;
  
  while (1)
    {
      s = select(socket+1,&r,NULL,&e,&tv);
      
      if (s==0)
        return 0;
  
      if (!FD_ISSET(socket,&r))
        return 0;

      n = read(socket,buf,count);
      
      if (memcmp(buf+6,macAddr,6))
        continue;  // not from the right box: count on fact that tv is time remaining
      
      if (buf[15] != index)
        continue;  // not intended for daemon
      
      return n;
    }
}

typedef struct 
{
  unsigned long macLoadAddress;
  unsigned long tcpLoadAddress;
  unsigned long startAddress;
  char *bootFile;
} tArchData;

// Whoever decided to use multiple formats to denote the EXACT
// SAME ADDRESS in the 8086 hub has some explaining to do... ;)

tArchData archData[] =
{
    {0,0,0,NULL},                                       // not used
    {0x00010000, 0x10000000, 0x10000000, "/etc/rpshsi.bin"},      // 8086 hubs
    {0x00100000, 0x00100000, 0x00100000, "/etc/rpshsi2p.bin"},    // original 2-port ARM
    {0x00000000, 0x00000000, 0x00000000, "/etc/devmast.bin"}      // deviceMaster
};

unsigned short IntelOrderS(unsigned short v)
{
  unsigned char t;
  union 
    {
      unsigned short s;
      unsigned char  b[2];
    }u;

  u.s = htons(v);
  t = u.b[1];
  u.b[1] = u.b[0];
  u.b[0] = t;
  return u.s;
}

unsigned long IntelOrderL(unsigned long v)
{
  union 
    {
      unsigned long l;
      unsigned char  b[4];
    }in,out;
  
  in.l = htonl(v);
  
  out.b[0] = in.b[3];
  out.b[1] = in.b[2];
  out.b[2] = in.b[1];
  out.b[3] = in.b[0];
  
  return out.l;
}


unsigned long GoofyOrderL(unsigned long v)
{
  union 
    {
      unsigned long l;
      unsigned char  b[4];
    }in,out;
  
  in.l = htonl(v);
  
  out.b[0] = in.b[1];
  out.b[1] = in.b[0];
  out.b[2] = in.b[3];
  out.b[3] = in.b[2];
  
  return out.l;
}

typedef struct
{
  tToolRequest tr __attribute__((packed));
  unsigned char lc[1460 - sizeof (tToolRequest)];
} __attribute__((packed)) tLoadRequest;


/*
 * returns: -2: hub was running other app -- so we sent reset command
 *          -1: failure -- couldn't talk to hub
 *           0: right application was already running
 *           1: we downlaoded application
 * 
 * Warning:  
 *
 * All commands have to go out in their own Ethernet packets.
 * 
 * That means you have to delay between write() calls on the
 * socket so that the network stack has time to send out the
 * previous command.  Othewise the stack might concatenate the
 * two write()s into a single Ethernet frame.
 * 
 * This is a bug in the protocol design.  We are trying to use
 * a stream service as a datagram service.  This is bad. We
 * shouldn't do this. 
 *
 */

int downloadTcp(int nslinkfd, int socket, struct si_box_config *config)
{
  unsigned char tcpBuf[sizeof (tLoadRequest)];
  tToolResponse *tr = (tToolResponse *)tcpBuf;
  tToolRequest  *trq = (tToolRequest *)tcpBuf;
  tLoadRequest  *lcreq = (tLoadRequest *)tcpBuf;
  int fd;
  int n;
  int ldAddr;
  int archId;
  unsigned char ldIndex;
  tArchData *a;
  tTcpTxQueueWait tqwait;

  tqwait.sock = socket;
  tqwait.msTimeout = 2000;
  
  trq->cmd = ID_TOOLREQ;
  trq->index = 0;
  trq->length = IntelOrderS(sizeof *trq);
  trq->addr = 0;
    
  write(socket,trq,sizeof *trq);
  
  if ((n=socketReadTimeout(socket,tcpBuf,sizeof tcpBuf, 200000)) <= 0)
    {
      syslog(LOG_ERR, "TCP ID request timeout");
      return -1;
    }

  if (n < sizeof (tOldToolResponse))
    {
      syslog(LOG_ERR, "TCP ID response too short");
      return -1;
    }
  
  if (n >= sizeof (tToolResponse))
    {
      tr->ProgramId = ntohs(tr->ProgramId);
      tr->ModelId   = ntohl(tr->ModelId);
      syslog(LOG_INFO, "TCP ID response from %02x:%02x:%02x:%02x:%02x:%02x %ld.%ld.%ld.%ld Model=%ld Arch=%d NumPorts=%d ProgId=%d BoardRev=%d",
             tr->MacAddr[0], tr->MacAddr[1], tr->MacAddr[2], tr->MacAddr[3], tr->MacAddr[4], tr->MacAddr[5],
             tr->IpAddr & 0xff, tr->IpAddr >> 8 & 0xff, tr->IpAddr >> 16 & 0xff, tr->IpAddr >> 24 & 0xff,
             tr->ModelId, tr->ArchId, tr->NumPorts, tr->ProgramId, tr->BoardRev);
    }
  else
    {
      // must be old box!
      syslog(LOG_INFO, "TCP ID response from %02x:%02x:%02x:%02x:%02x:%02x %ld.%ld.%ld.%ld %08lx",
             tr->MacAddr[0], tr->MacAddr[1], tr->MacAddr[2], tr->MacAddr[3], tr->MacAddr[4], tr->MacAddr[5],
             tr->IpAddr & 0xff, tr->IpAddr >> 8 & 0xff, tr->IpAddr >> 16 & 0xff, tr->IpAddr >> 24 & 0xff,
             tr->AvailMemAddr);

      // we have to look at memory values to figure out who we're talking to :(

      if (tr->MacAddr[3] == 0x10 || tr->MacAddr[3] == 0x11)
        {
          tr->ArchId = 2;         // old ARM 2-port or IA
          if (tr->AvailMemAddr == 0x1000)
            tr->ProgramId = 0;      // bootloader
          else
            tr->ProgramId = 1;      // RPSH-Si app
        }
      else if (tr->MacAddr[3] == 0x06)
        {
          tr->ArchId = 1;         // 8086
          if (tr->AvailMemAddr == 0x10000000)
            tr->ProgramId = 0;      // bootloader
          else
            tr->ProgramId = 1;      // RPSH-Si app
        }
      else
        {
          tr->ArchId = 0;         // don't know
          tr->ProgramId = 0xff;   // not a bootloader
        }
    }

  config->arch_id = tr->ArchId;
  
  if (tr->ProgramId == 1)
    return 0;                     // right app is already running
  
  if (tr->ProgramId != 0)
    {
      // wrong app running, try sending reset command
      syslog(LOG_INFO, "Resetting device at %ld.%ld.%ld.%ld\n",
             tr->IpAddr & 0xff, tr->IpAddr >> 8 & 0xff, tr->IpAddr >> 16 & 0xff, tr->IpAddr >> 24 & 0xff);
      trq->cmd = RESET_TOOLREQ;
      trq->index = 0;
      trq->length = IntelOrderS(sizeof *trq);
      trq->addr = 0;
      write(socket,trq,sizeof *trq);
      if (ioctl(nslinkfd,RPSHSI_TCP_TXQUEUE_WAIT,&tqwait))
        syslog(LOG_ERR,"txqueue_wait %s",strerror(errno));
      return -2;
    }
  
  if (tr->ArchId <= 0 || tr->ArchId >= ((sizeof archData)/(sizeof archData[0])))
    {
      syslog(LOG_ERR, "Unknown architecture ID: %d", tr->ArchId);
      return -1;
    }
  
  archId = tr->ArchId;
  
  a = archData + archId;
  fd = open(a->bootFile,O_RDONLY);

  if (fd < 0)
    {
      syslog(LOG_ERR,"Can't open boot file %s", a->bootFile);
      return -1;
    }

  // we're going to re-use tcpBuf (over-writing id response data)

  ldAddr = a->tcpLoadAddress;
  ldIndex = 0;
  
  while (1)
    {
      int len;
      lcreq->tr.cmd = LOAD_TOOLREQ;
      lcreq->tr.index = ldIndex;
      lcreq->tr.addr = IntelOrderL(ldAddr);
      n = read(fd,lcreq->lc,(sizeof lcreq->lc) - 200);
      if (n <= 0)
        break;
      len = n + sizeof (tToolRequest);
      lcreq->tr.length = len;
      if (len != write(socket, lcreq, len))
        {
          syslog(LOG_ERR,"Error writing to socket");
          return -1;
        }
      
      usleep(5000);  // 5ms delay to let packet get sent
      
      // wait for data to get acked (only needed for pre-DM hubs)

      if (archId != 3)
        if (ioctl(nslinkfd,RPSHSI_TCP_TXQUEUE_WAIT,&tqwait))
          syslog(LOG_ERR,"txqueue_wait %s",strerror(errno));

      ++ldIndex;
      ldAddr += n;
    }
  
  trq->cmd = GOTO_TOOLREQ;
  trq->index = 0;
  trq->length = IntelOrderS(sizeof *trq);
  trq->addr = IntelOrderL(a->startAddress);

  write(socket,trq, sizeof *trq);
  if (archId != 3)
    if (ioctl(nslinkfd,RPSHSI_TCP_TXQUEUE_WAIT,&tqwait))
      syslog(LOG_ERR,"txqueue_wait %s",strerror(errno));
  
  return 1;
}


void setup_tcp_dev(int fd,  struct si_box_config *config, int boxIndex)
{
  int sk, rc;
  struct sockaddr_in sockaddr;
  struct si_tcp_set tcpset;
  int resetRetryCount = -1;
  int retryCount = 0;

  tcpset.sk = 0;
  tcpset.rbuf = NULL;
  tcpset.rbuf_size = 0;
  tcpset.wbuf = NULL;
  tcpset.wbuf_size = 0;

  while (1)
    {
      int n;
      
      if (retryCount < 6)
        ++retryCount;
      
      if (retryCount > 0)
        sleep(1<<retryCount);
      
      sockaddr.sin_family = AF_INET;
      sockaddr.sin_port = htons(0x11fe);
      sockaddr.sin_addr.s_addr = config->ip_addr;
      //syslog(LOG_INFO, "Child %d creating TCP socket\n", boxIndex);
      sk = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
      if (sk < 0)
        {
          syslog(LOG_ERR, "Could not create socket for NS-Link device.");
          exit(1);
        }
      
      child_socket = sk;
      
      //syslog(LOG_INFO, "Child %d connecting socket %d\n", boxIndex, sk);
      rc = connect(sk, (struct sockaddr*)&sockaddr, sizeof(struct sockaddr_in));
      
      if (rc < 0)
        {
          syslog(LOG_ERR, "Could not connect to NS-Link device at %s",inet_ntoa(sockaddr.sin_addr));
          close(sk);
          continue;
        }
          
      // download executable binary file to hub and restart it

      n = downloadTcp(fd,sk,config);
      
      if (n == -2 && retryCount < 2)
        {
          // box was running other app and has been reset
          ++resetRetryCount;
          close(sk);
          sleep(3);
          continue;
        }
      
      if (n < 0)
        {
          syslog(LOG_ERR,"Couldn't download to NS-Link device, %s.",inet_ntoa(sockaddr.sin_addr));
          close(sk);
          continue;
        }
      
      // if download returned 0, then hub was already loaded, otherwise
      // it has just been restarted -- so we have to close the socket,
      // wait for the thing to re-start, and open a new socket
          
      if (n != 0)
        {
          //syslog(LOG_INFO, "Child %d closing socket %d\n",boxIndex, sk);
          
          if (close(sk))
            syslog(LOG_INFO, "Child %d error closing socket %d: '%s'\n",boxIndex,sk,strerror(errno));
          sleep(3);
          
          //syslog(LOG_INFO, "Child %d creating socket\n", boxIndex);
              
          sk = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
          if (sk < 0)
            {
              syslog(LOG_ERR, "Could not create socket for NS-Link device.");
              continue;
            }
          child_socket = sk;
          
          //syslog(LOG_INFO, "Child %d connecting socket\n", boxIndex);
          rc = connect(sk, (struct sockaddr*)&sockaddr, sizeof(struct sockaddr_in));
          if (rc < 0)
            {
              syslog(LOG_ERR, "Could not connect to NS-Link device at %s",inet_ntoa(sockaddr.sin_addr));
              close(sk);
              continue;
            }
        }
      
      // hub should be up and running, turn the driver loose on it.

      tcpset.boxid = boxIndex;
      tcpset.sk = sk;
      tcpset.rbuf = malloc(4096);
      tcpset.rbuf_size = 4096;
      tcpset.wbuf = malloc(1460);
      tcpset.wbuf_size = 1460;
      
      retryCount = 0;
      
      if ((tcpset.rbuf == NULL) || (tcpset.wbuf == NULL))
        {
          syslog(LOG_ERR, "Could not allocate memory for NS-Link device, %s.", inet_ntoa(sockaddr.sin_addr));
          exit(1);
        }
 
      // syslog(LOG_INFO,"Child %d configuring hub\n",boxIndex);
      
      config_rpshsi(fd, config, boxIndex);

      // syslog(LOG_INFO,"Child %d setting TCP\n",boxIndex);

      rc = ioctl(fd, RPSHSI_SET_TCP, &tcpset);
      
      if (rc)
        {
          syslog(LOG_ERR, "Could not find driver, %s.", inet_ntoa(sockaddr.sin_addr));
          exit(1);
        }
          
      // syslog(LOG_INFO,"Child %d shutting down TCP\n",i);
      rc = shutdown(sk, 2);
      rc = close(sk);
      free(tcpset.wbuf);
      free(tcpset.rbuf);
    }
}

char *mactoa(unsigned char *m)
{
  static char buf[20];
  sprintf(buf,"%02x:%02x:%02x:%02x:%02x:%02x",m[0],m[1],m[2],m[3],m[4],m[5]);
  return buf;
}

typedef struct
{
  unsigned char   destMac[6];
  unsigned char   srcMac[6];
  unsigned short  proto;
  unsigned char   prodId;
  unsigned char   concId;
  unsigned char   packetClass;
  unsigned short  length;
  unsigned char   adminCmd;
  unsigned char   adminData[1470];
} __attribute__((packed))tAdminPacket;

#define AdminHeaderLen 20

typedef struct 
{
  unsigned char  tilde;
  unsigned short length;
  unsigned short cmd;
  unsigned long  offset;
  unsigned char  data[1460];
}__attribute__((packed)) tMacBootPacket;

typedef struct
{
  unsigned char  tilde;
  unsigned short length;
  unsigned short cmd;
  unsigned long  ioctl;
  unsigned long  startAddr;
  unsigned char checksum;
}__attribute__((packed)) tMacGoPacket;

typedef struct
{
  unsigned char    macAddr[6];
  unsigned char    flags;
  unsigned char    unused0;
  unsigned char    unused1;
  unsigned char    unused2;
  unsigned long    modelId;
  unsigned char    archId;
  unsigned char    numPorts;
  unsigned short   progId;
  unsigned char    boardRev;
  unsigned char    bootVersMajor;
  unsigned char    bootVersMinor;
  unsigned char    numAppRecords;
} __attribute__((packed)) tMacIdResp;

typedef struct
{
  unsigned char    macAddr[6];
  unsigned char    flags;
  unsigned char    unused0;
  unsigned char    unused1;
  unsigned char    unused2;
} __attribute__((packed)) tOldMacIdResp;

#define ADMIN_BOOT           0
#define ADMIN_BOOT_PACKET    1
#define ADMIN_ID_QUERY       2
#define ADMIN_ID_REPLY       3
#define ADMIN_LOOP           4
#define ADMIN_RESET          5
#define ADMIN_TRACE          6
#define ADMIN_INIT           7
#define ADMIN_READMEM        8
#define ADMIN_WRITEMEM       9
#define ADMIN_TEST           10
#define ADMIN_DATA           11
#define ADMIN_ELBACK         12
#define ADMIN_READIO         13
#define ADMIN_WRITEIO        14
#define ADMIN_RSMODE         15
#define ADMIN_COMPMEM        16
#define ADMIN_MOVEMEM        17
#define ADMIN_TIMXMIT        18

#define ADMIN_IPINFO_SET    32
#define ADMIN_IPINFO_GET    33
#define ADMIN_IPINFO_ERASE  34

/*
 * returns: -1: failure
 *           0: right application was already running
 *           1: we downlaoded application
 */

int downloadMac(int socket, char *destMac, char *myMac, int boxIndex, struct si_box_config *config)
{
  tAdminPacket  cmd;
  tAdminPacket  rsp;
  tMacIdResp    *id =  (tMacIdResp*)&rsp.adminData;
  tMacBootPacket *bp = (tMacBootPacket*)&cmd.adminData;
  tMacGoPacket   *gp = (tMacGoPacket*)&cmd.adminData;
  int n,fd;
  int ldAddr,len;
  unsigned char ldIndex;
  tArchData *a;
  unsigned char *cp;
  unsigned char checksum;
  int retryCount = 0;

  // stuff common to all commands we will send

  memcpy(cmd.destMac,destMac,6);
  memcpy(cmd.srcMac,myMac,6);
  cmd.proto = htons(0x11fe);
  cmd.prodId = 0xaa;
  cmd.concId = 0x80 + boxIndex;
  cmd.packetClass = 1;

retry:  
  // send ID request
          
  cmd.adminCmd = ADMIN_ID_QUERY;
  memcpy(cmd.adminData,myMac,6);
  cmd.adminData[6] = 0;    // flags (bit0 == 0 ==> request ownership)
  cmd.length = IntelOrderS(8);
  write(socket,&cmd,AdminHeaderLen+8);
  
  if ((n=packetReadTimeout(socket, destMac, cmd.concId, &rsp, sizeof rsp, 200000)) <= 0)
    {
      syslog(LOG_ERR, "MAC ID request timeout");
      return -1;
    }

  if (n < AdminHeaderLen + sizeof (tOldMacIdResp) || IntelOrderS(rsp.length) < 7)
    {
      syslog(LOG_ERR, "MAC ID response too short");
      return -1;
    }

  if (rsp.adminCmd != ADMIN_ID_REPLY)
    {
      syslog(LOG_ERR, "MAC ID response has wrong cmd type");
      return -1;
    }

  
  id->progId = ntohs(id->progId);
  id->modelId = ntohl(id->modelId);
  
  if (n >= sizeof (tMacIdResp) && 
      IntelOrderS(rsp.length) >= 22 &&
      id->macAddr[3] != 0x06)
    {
      syslog(LOG_INFO, "MAC ID response from %s flags=%02x, Model=%ld Arch=%d NumPorts=%d ProgId=%d BoardRev=%d",
             mactoa(id->macAddr),id->flags, id->modelId, id->archId, id->numPorts, id->progId, id->boardRev);
    }
  else
    {
      // must be old box!
      syslog(LOG_INFO, "MAC ID response from %s flags=%02x\n", mactoa(id->macAddr),id->flags);

      // we have to look at addr and flag values to figure out who we're talking to :(

      if ((id->macAddr[3] == 0x10 || id->macAddr[3] == 0x11))
        id->archId = 2;
      else if (id->macAddr[3] == 0x06)
        id->archId = 1;
      else
        id->archId = 0;         // don't know

      if (id->flags & 0x01)
        id->progId = 1;
      else
        id->progId = 0;
    }

  config->arch_id = id->archId;
  
  if (id->progId == 1)
    return 0;                     // right app running

  
  if (id->progId != 0)
    {
      // wrong app running, try sending reset command

      if (retryCount>2)
        {
          syslog(LOG_ERR, "Device at %s failed to reset\n",mactoa(id->macAddr));
          return -1;
        }
      syslog(LOG_INFO,"Resetting device at %s\n",mactoa(id->macAddr));
      cmd.adminCmd = ADMIN_RESET;
      cmd.adminData[0] = 0x55;
      cmd.adminData[1] = 0x55;
      cmd.length = IntelOrderS(2);
      write(socket,&cmd,AdminHeaderLen+2);
      sleep(3);
      ++retryCount;
      goto retry;
    }
  
  // it's a bootloader we're talking to...
  
  if (id->flags & 0x02)
    {
      // already owned: somebody else is downloading
      syslog(LOG_INFO, "Somebody else is downloading...");
      sleep(1);
      goto retry;
    }
  
  if (id->archId <= 0 || id->archId >= ((sizeof archData)/(sizeof archData[0])))
    {
      syslog(LOG_ERR, "Unknown architecture ID: %d", id->archId);
      return -1;
    }

  a = archData + id->archId;
  
  fd = open(a->bootFile,O_RDONLY);

  if (fd < 0)
    {
      syslog(LOG_ERR,"Can't open boot file %s", a->bootFile);
      return -1;
    }

  ldAddr = a->macLoadAddress;
  ldIndex = 0;
  
  while (1)
    {
      cmd.adminCmd = ADMIN_BOOT_PACKET;
      bp->tilde = '~';
      bp->length = 0;   // fill in later
      bp->cmd    = IntelOrderS(8);
      bp->offset = IntelOrderL(ldAddr);

      n = read(fd,bp->data,sizeof bp->data);
      
      if (n <= 0)
        break;
      
      len = n + 6;
      
      bp->length = IntelOrderS(len);
      cmd.length = IntelOrderS(len+14);
      
      checksum = 0;
      for (cp=(unsigned char*)&bp->length; cp < bp->data+n; ++cp)
        checksum += *cp;
          
      bp->data[n] = ~checksum;
      
      len = 1 + (&(bp->data[n]) - (unsigned char*)&cmd);
      
      if (len != write(socket, &cmd, len))
        {
          syslog(LOG_ERR,"Error writing to socket");
          return -1;
        }
      
      ldAddr += n;
      
      if ((n=packetReadTimeout(socket, destMac, cmd.concId, &rsp, sizeof rsp, 200000)) <= 0)
        {
          syslog(LOG_ERR, "download response timeout");
          return -1;
        }

      if (n < 24)
        {
          syslog(LOG_ERR, "download response too short");
          return -1;
        }

      if (rsp.adminCmd != ADMIN_BOOT_PACKET ||
          rsp.adminData[0] != '|' ||
          rsp.adminData[1] != 0x02 ||
          rsp.adminData[2] != 0x00 ||
          rsp.adminData[3] != 0x88)
        {
          syslog(LOG_ERR, "download response invalid");
          syslog(LOG_ERR, "%02x: %02x %02x %02x %02x\n",
                 rsp.adminCmd,
                 rsp.adminData[0],
                 rsp.adminData[1],
                 rsp.adminData[2],
                 rsp.adminData[3]);
          return -1;
        }
    }

  // send "go" command

  gp->tilde  = '~';
  gp->length = IntelOrderS(10);
  gp->cmd    = 5;
  gp->ioctl  = IntelOrderL(12);
  gp->startAddr = GoofyOrderL(a->startAddress);

  checksum = 0;
  for (cp = (unsigned char*)&gp->length; cp < &gp->checksum; ++cp)
    checksum += *cp;
  
  gp->checksum = ~checksum;
  
  len = 1 + (&gp->checksum - (unsigned char*)&cmd);
  
  write(socket, &cmd, len);
  
  return 1;
}


void setup_mac_dev(int fd,   struct si_box_config *config, int boxIndex)
{
  int sk, rc;
  struct sockaddr_ll sockaddr;
  struct ifreq ifr;
  struct si_mac_set macset;
  char myMac[6];
  int retryCount = -1;

  while (1)
    {
      int n;
      
      if (retryCount < 6)
        ++retryCount;
      
      if (retryCount > 0)
        sleep(1<<retryCount);
      
      //syslog(LOG_INFO, "Child %d creating packet socket\n", boxIndex);
      sk = socket(AF_PACKET, SOCK_RAW, htons(0x11fe));
      
      if (sk < 0)
        {
          syslog(LOG_ERR, "Could not create socket for NS-Link device.");
          exit(1);
        }

      // look up interface index (needed for bind())

      strncpy(ifr.ifr_name, config->interface, sizeof(ifr.ifr_name));
      ifr.ifr_name[(sizeof(ifr.ifr_name))-1] = '\0';
      if (ioctl(sk, SIOCGIFINDEX, &ifr))
        {
          syslog(LOG_ERR,"Child %d couldn't find interface '%s'\n",boxIndex,config->interface);
          exit(1);
        }
      

      sockaddr.sll_family = AF_PACKET;
      sockaddr.sll_protocol = htons(0x11fe);
      sockaddr.sll_ifindex = ifr.ifr_ifindex;

      // only want to see packets from specified ethernet interface
      
      if (bind(sk,(struct sockaddr*) &sockaddr, sizeof sockaddr))
        {
          syslog(LOG_ERR,"Child %d couldn't bind packet socket to %s\n",boxIndex,config->interface);
          exit(1);
        }
      
      child_socket = sk;

      n = sizeof sockaddr;
      getsockname(sk, (struct sockaddr*) &sockaddr, &n);
    
      memcpy(myMac,sockaddr.sll_addr,6);

      //syslog(LOG_INFO, "Child %d bound socket to %s (%s)\n", boxIndex, config->interface,mactoa(myMac));
          
      // download executable binary file to hub and restart it

      n = downloadMac(sk,config->ether_addr,myMac,boxIndex,config);
      
      close(sk);
      
      if (n < 0)
        {
          syslog(LOG_ERR,"Couldn't download to NS-Link device, %s.", mactoa(config->ether_addr));
          continue;
        }

      if (n > 0)
        sleep(3);  // just started up, so let the thing wake up

      // hub should be up and running, turn the driver loose on it.

      // syslog(LOG_INFO,"Child %d configuring hub\n",boxIndex);

      config_rpshsi(fd, config, boxIndex);
  
      
      // syslog(LOG_INFO,"Child %d setting Mac\n",boxIndex);

      retryCount = 0;
      macset.boxid = boxIndex;
      rc = ioctl(fd, RPSHSI_SET_MAC, &macset);
      
      if (rc)
        {
          syslog(LOG_ERR, "Could not set device, %s.", mactoa(config->ether_addr));
          exit(1);
        }
    }
}


void setup_devices(int fd)
{
  int i;
  int newpid;
  struct si_box_config *config;
  
  for (i=0; i< num_box_config; ++i)
    {
      config = box_config + i;
      
      newpid = fork();
      
      if (newpid)
        {
          /* parent */
          child_pids[i] = newpid;
          continue;
        }
      
      /* child */
      
      if (config->ip_addr)
        setup_tcp_dev(fd,config,i);
      else
        setup_mac_dev(fd,config,i);
    }
}


int main(int argc, char **argv)
{
  int major, fd;

  if (getuid())
    {
      fprintf(stderr, "You must be root to run NS-Link daemon!\n");
      exit(1);
    }

  openlog("nslinkd", LOG_NDELAY | LOG_PID, LOG_DAEMON);
  syslog(LOG_INFO, "Starting NS-Link daemon, version %s (%s)",SI_VERSION, SI_DATE);
  syslog(LOG_INFO, "Built %s %s\n",__DATE__,__TIME__);
  
  parse_config("/etc/nslink.conf");

  if (boot_file4_8)
    archData[1].bootFile = boot_file4_8;
  if (boot_file2)
    archData[2].bootFile = boot_file2;
  if (boot_fileDM)
    archData[3].bootFile = boot_fileDM;

  major = get_major_number("NSLinkctl");

  if (major < 0)
    {
      syslog(LOG_ERR, "Couldn't load control file");
      fprintf(stderr, "Couldn't find control file.  %s\n", not_loaded);
      exit(1);
    }
  fd = open_dev(major << 8, O_RDWR);
  if (fd < 0)
    {
      syslog(LOG_ERR, "Couldn't open control file");
      fprintf(stderr, "Couldn't open control device.  %s\n", not_loaded);
      exit(1);
    }
  check_pid_file();
  if (become_daemon() < 0)
    {
      syslog(LOG_ERR, "Couldn't become daemon");
      fprintf(stderr, "Error forking child process while "
	      "starting daemon mode\n");
      exit(1);
    }
  put_pid_file();
  signal(SIGHUP, sighup_handler);
  signal(SIGTERM, sigterm_handler);
  signal(SIGINT, sigterm_handler);
  
  create_devices();

  setup_devices(fd);

  pause();
  return 0;
}
