#!/bin/sh
# This script is responsible for installing the NS-Link driver

if [ `whoami` != root ]; then
  echo "You must be root to install the device driver!"
  exit 1
fi

# Check to see whether or not mod_rocket.o was built; if it wasn't built,
# we don't need to install it.

if [ -f mod_nslink.o ]; then
  OBJS="mod_nslink.o nslink.o"
else
  OBJS=nslink.o
fi

# Install a copy of the driver in the version independent location
if [ ! -d /lib/modules ]; then 
  mkdir -p /lib/modules
fi

cp $OBJS /lib/modules

# Install a copy of the driver in a kernel version specific location
if [ -f .kver ]; then
  KVER=`cat .kver`;
else
  KVER=`grep UTS_RELEASE /usr/include/linux/version.h | \
  awk '{print $3}' | tr -d \"`
fi
RVER=`uname -r`
echo " "
echo "Installing NS-Link driver for Linux $KVER"
echo " "
if [ "$KVER" != "$RVER" ]; then
  echo "WARNING: The current kernel is actually version $RVER."
  echo " "
fi

# Actually install the kernel modules

if [ ! -d /lib/modules/$KVER/misc ]; then 
  mkdir -p /lib/modules/$KVER/misc
fi

cp $OBJS /lib/modules/$KVER/misc

# If the modules version wasn't built, remove it from the install tree, 
# lest we get led astray

if [ ! -f mod_nslink.o ]; then
  rm -f /lib/modules/mod_nslink.o /lib/modules/$KVER/misc/mod_nslink.o
fi

# install admin utilities

install --mode=700 nslinkadmin      /usr/sbin
install --mode=700 nslinktool       /usr/sbin
install --mode=700 nslinkrelease    /usr/sbin
install --mode=700 nslinkrelease.py /usr/sbin

#install man pages
for f in *.8
  do
  install --mode=644 -D $f /usr/local/man/man8/$f
  done
for f in *.5
  do
  install --mode=644 -D $f /usr/local/man/man5/$f
  done

# Install the nslink microcode files

cp rpshsi.bin   /etc/rpshsi.bin
cp rpshsi2p.bin /etc/rpshsi2p.bin
cp devmast.bin  /etc/devmast.bin

# Install nslinkd

if [ -f /sbin/nslinkd ] ; then
    mv /sbin/nslinkd /sbin/nslinkd.old
fi
cp nslinkd /sbin/nslinkd
strip /sbin/nslinkd
chmod 755 /sbin/nslinkd
rm -f /sbin/nslinkd.old

# Copy in the nslink.conf file, if isn't already installed
if [ -f /etc/rpshsi.conf ]; then
  echo "Using existing /etc/rpshsi.conf -- renaming to nslink.conf"
  mv /etc/rpshsi.conf /etc/nslink.conf
elif [ ! -f /etc/nslink.conf ]; then
  cp nslink.conf /etc
  echo "Make sure you edit the /etc/nslink.conf file!"
else
  echo "Using existing /etc/nslink.conf"
fi

# Now we install the boot script 

if [ -d /etc/rc.d/init.d ]; then 
  rm -f /etc/rc.d/init.d/rpshsi
  cp rc.nslink /etc/rc.d/init.d/nslink
elif [ -d /etc/init.d ]; then 
  rm -f /etc/init.d/rpshsi
  cp rc.nslink /etc/init.d/nslink
else
  rm -f /etc/rc.d/rc.rpshsi
  if [ ! -f /etc/rc.d/rc.nslink ]; then
    if [ ! -d /etc/rc.d ]; then mkdir -p /etc/rc.d; fi
    cp rc.nslink /etc/rc.d
    echo "The initialization script has been installed in "
    echo "/etc/rc.d/rc.nslink.  You will need to edit your "
    echo "/etc/rc files to run /etc/rc.d/rc.nslink"
  fi
  # don't know where to put symlinks, so give up
  exit 0
fi

# remove old rpshsi symlinks

for f in `find /etc -name "[SK][0-9][0-9]rpshsi"`
 do
 rm -f $f
 done
 

# Set up symlinks with chkconfig if we can

if [ -f /sbin/chkconfig ]; then
  /sbin/chkconfig --add nslink
  echo "NS-Link chkconfig (startup) settings:"
  /sbin/chkconfig --list nslink
  exit 0
fi

# chkconfig not found, so try to do things manually

if [ -d /etc/rc.d/init.d ]; then 
  # This is a RedHat or other system using a System V init scheme
  if [ -s /etc/rc.d/rc2.d/S95nslink -o -s /etc/rc.d/rc2.d/K05nslink ]; then
    exit 0
  fi
  
  echo "You are using a system which uses the System V init"
  echo "scheme.  The initialization script has been installed"
  echo "in /etc/rc.d/init.d and it has been automatically "
  echo "set up to be run when you reboot"

  ln -sf ../init.d/nslink /etc/rc.d/rc2.d/S95nslink
  ln -sf ../init.d/nslink /etc/rc.d/rc3.d/S95nslink
  ln -sf ../init.d/nslink /etc/rc.d/rc5.d/S95nslink
  ln -sf ../init.d/nslink /etc/rc.d/rc0.d/K05nslink
  ln -sf ../init.d/nslink /etc/rc.d/rc1.d/K05nslink
  ln -sf ../init.d/nslink /etc/rc.d/rc6.d/K05nslink
elif [ -d /etc/init.d ]; then 
  # This is a Debian system
  if [ -s /etc/rc2.d/S95nslink -o -s /etc/rc2.d/K05nslink ]; then
    exit 0
  fi
    
  echo "You are using a system which uses the Debian "
  echo "System V init directory layout.  The initialization"
  echo "script has been installed in /etc/init.d and it"
  echo "has been automatically set up to be run when you reboot"

  ln -sf ../init.d/nslink /etc/rc2.d/S95nslink
  ln -sf ../init.d/nslink /etc/rc3.d/S95nslink
  ln -sf ../init.d/nslink /etc/rc5.d/S95nslink
  ln -sf ../init.d/nslink /etc/rc0.d/K05nslink
  ln -sf ../init.d/nslink /etc/rc1.d/K05nslink
  ln -sf ../init.d/nslink /etc/rc6.d/K05nslink
fi
